from abc import ABC, abstractmethod
from KnotenUndKantenSymbole import KnotenSymbol, KantenSymbol

from intern.zeichenfenster import *


class Woerterbucheintrag:
    """
    Die Klasse Woerterbucheintrag dient zur Speicherung von
    Wörterbucheinträgen fuer ein Englisch-Deutsches
    Wörterbuch. Dabei kann jedem englischen Wort mehrere
    deutsche Bedeutungen zugeordnet werden, wobei alle Bedeutungen in
    einer einzigen Zeichenkette Bedeutung gespeichert werden.
    """
    def __init__(self, wortNeu, bedeutungNeu):
        """
        Konstruktor fuer Objekte der Klasse Woerterbucheintrag
        :param wortNeu: das englische Wort.
        :param bedeutungNeu: die deutsche(n) Bedeutung(en).
        """
        self.wort = wortNeu
        self.bedeutung = bedeutungNeu

    def InformationAusgeben(self):
        """
        Gibt Information ueber das Datenelement zu Kontrollzwecken
        auf das Terminalfenster aus.
        """
        print(f"{self.wort}: {self.bedeutung}")

    def IstGleich(self, dvergleich):
        """
        Vergleicht zwei Datenelemente auf Gleichheit.
        :param dvergleich: Datenelement mit dem das Objekt verglichen wird.
        :return: True, wenn die beiden Datenelemente gleichen Schluessel haben.
        """
        return self.wort == dvergleich.WortGeben()

    def IstGroesserAls(self, dvergleich):
        """
        Vergleicht zwei Datenelemente bezueglich der Ordnungsrelation.
        :param vergleichselement: Datenelement mit dem das Objekt verglichen wird.
        :return: True, wenn das aktuelle Element einen größeren Schluessel hat, als das angegebene Vergleichselement.
        """
        return self.wort > dvergleich.WortGeben()

    def SchluesselIstGleich(self, vergleichsSchluessel):
        """
        Vergleicht zwei Schluessel, die als Zeichenketten vorliegen, auf Gleichheit.
        :param vergleichsSchluessel: Schluessel mit dem der Schluessel des Objekt verglichen wird.
        :return: True, wenn die beiden Schluessel gleich sind.
        """
        return self.wort == vergleichsSchluessel

    def SchluesselIstGroesserAls(self, vergleichsSchluessel):
        """
        Vergleicht zwei Schluessel, die als Zeichenketten vorliegen, bezueglich der Ordnungsrelation.
        :param vergleichsSchluessel: Schluessel mit dem der Schluessel des Objekt verglichen wird.
        :return: True, wenn das aktuelle Element einen größeren Schluessel hat, als das angegebene Vergleichselement.
        """
        return self.wort > vergleichsSchluessel

    def SchluesselIstKleinerAls(self, vergleichsSchluessel):
        """
        Vergleicht zwei Schluessel, die als Zeichenketten vorliegen, bezueglich der Ordnungsrelation.
        :param vergleichsSchluessel: Schluessel mit dem der Schluessel des Objekt verglichen wird.
        :return: True, wenn das aktuelle Element einen kleineren Schluessel 
        hat, als das angegebene Vergleichselement.
        """
        return self.wort < vergleichsSchluessel

    def WortGeben(self):
        """
        Geben-Methode zum Attribut wort
        :return: das englische Wort
        """
        return self.wort

    def BedeutungGeben(self):
        """
        Geben-Methode zum Attribut bedeutung
        :return: bedeutung: die deutsche(n) Bedeutung(en) als eine Zeichenkette
        """
        return self.bedeutung

    def BedeutungSetzen(self, bedeutungNeu):
        """
        Setzen-Methode zum Attribut bedeutung
        :param bedeutungNeu: neue Bedeutung fuer das englische Wort
        """
        self.bedeutung = bedeutungNeu


class Baumelement(ABC):
    """
    Abstrakte Klasse Baumelement zur Realisierung einer Baumstruktur mit Hilfe des Entwurfsmusters
    Kompositum: Die Klasse legt ausschließlich die Schnittstelle fuer die Klassen
    Knoten und Abschluss in Form von Methoden fest.
    """
    def __init__(self):
        """
        Der Konstruktor ist nur der Vollständigkeit halber angegeben.
        """
        pass  # nichts zu tun

    @abstractmethod
    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement, falls sein Schluessel noch nicht vorhanden ist,  
        in die bestehende Datenstruktur ein. D.h. die Eigenschaft geordneter Binärbaum 
        bleibt erhalten.
        :param datenNeu: das Datenelement des einzufuegenden Knoten
        :return: (neues) Nachfolgeelement des Aufrufers, None falls es kein Datenelement mit dem eingegebenen Schluessel gibt
        """
        pass

    @abstractmethod
    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel identifiziert wird
        und gibt dieses aus.
        :param suchSchluessel: Schluessel nach dem gesucht wird als Zeichenkette 
        :return: gesuchtes Datenelement bzw. None, falls die Suche erfolglos ist
        """
        pass

    @abstractmethod
    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist. 
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: True, im Erfolgsfall, False sonst.
        """
        pass

    @abstractmethod
    def HöheGeben(self, aktTiefe):
        """
        Berechnet die Höhe des Baums und gibt diesen Wert zurueck
        :param aktTiefe: Tiefe des aktuellen Knotens
        :return: die Höhe = maximale Tiefe (der bisher besuchten Knoten)
        """
        pass

    @abstractmethod
    def TiefeGeben(self, suchSchluessel, tiefe):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette
        :param tiefe: die Tiefe des aktuellen Knotens
        :return: Tiefe als positive Zahl, bzw. -1 falls der suchSchluessel nicht vorhanden ist.
        """
        pass


class Abschluss(Baumelement):
    """
    Klasse Abschluss als Bestandteil eines geordneten Binärbaums.
    (Implementierung mit dem Entwurfsmuster Kompositum)
    """
    def __init__(self):
        super().__init__()
        self.knotenSymbol = None
        self.kantenSymbol = None

    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement in die bestehende Datenstruktur ein, indem es einen neuen Knoten 
        (mit zwei Abschluss-Objekten als Nachfolger) erzeugt und eine Referenz auf diesen Knoten
        zurueckgibt. Das Abschluss-Objekt, dass diese Methode ausfuehrt, wird dabei aus dem Baum entfernt.
        :param datenNeu: einzufuegendes Datenelement 
        :return: Referenz auf den neu erzeugten Knoten
        """
        if self.knotenSymbol:
            self.knotenSymbol.Entfernen()
        if self.kantenSymbol:
            self.kantenSymbol.Entfernen()
        return Knoten(datenNeu)

    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel identifiziert wird
        und gibt dieses aus.
        :param suchSchluessel: Schluessel nach dem gesucht wird als Zeichenkette 
        :return: None, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return None

    def HöheGeben(self, aktTiefe):
        """
        Berechnet die Höhe des Baums und gibt diesen Wert zurueck
        :param aktTiefe: Tiefe des aktuellen Knotens
        :return: die uebergebene Höhe -1, da hier beim Abschluss das Ende eines Astes erreicht ist und 
        die Abschlusselemente bei der Höhenberechnung nicht einbezogen werden.
        """
        return aktTiefe - 1

    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist. 
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: False, da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return False

    def TiefeGeben(self, suchSchluessel, tiefe):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette
        :param tiefe: aktuelle Tiefe im Baum
        :return: -1 da die Suche erfolglos ist (Abschluss-Objekt kennzeichnet den Ende eines Astes!).
        """
        return -1
    
    def KnotenZeichnen(self, xmin, xmax, y, dy, symbolVorgänger):
    
        knotenSymbol = KnotenSymbol(x=0, y=0, radius=8, farbe="weiss", bezeichner="Abschluss" )
        knotenSymbol.PositionSetzen(xmin + (xmax - xmin) / 2, y);
        # Eine Kante wird nur gezeichnet, wenn es einen Vorgänger gibt (d.h. dieser Knoten nicht die Wurzel ist)
        if symbolVorgänger != None:
            kantenSymbol =  KantenSymbol(symbolVorgänger, knotenSymbol, False, "", 3, "schwarz");            


class Knoten(Baumelement):
    """
    Klasse Knoten als Bestandteil eines geordneten Binärbaums.
    (Implementierung mit Entwurfsmuster Kompositum)
    """
    def __init__(self, datenNeu, linkerNachfolgerNeu=None, rechterNachfolgerNeu=None):
        """
        Konstruktor fuer Objekte der Klasse Knoten ohne linken und rechten Nachfolger, d.h. es werden
        zwei Abschluss-Objekte erzeugt und den entsprechenden Referenzattributen zugewiesen.
        :param datenNeu: Referenz auf ein Objekt der Klasse Woerterbucheintrag, das vom Knoten verwaltet wird.
        :param linkerNachfolgerNeu: der linke Nachfolger des Knotens (default ist ein neuer Abschluss)
        :param rechterNachfolgerNeu: der rechte Nachfolger des Knotens (default ist ein neuer Abschluss)
        """
        super().__init__()
        self.daten = datenNeu
        self.linkerNachfolger = linkerNachfolgerNeu if linkerNachfolgerNeu else Abschluss()
        self.rechterNachfolger = rechterNachfolgerNeu if rechterNachfolgerNeu else Abschluss()

    def Einfuegen(self, datenNeu):
        if self.daten.IstGleich(datenNeu):
            print("Die Daten existieren schon!")
        elif self.daten.IstGroesserAls(datenNeu):
            self.linkerNachfolger = self.linkerNachfolger.Einfuegen(datenNeu)
        else:
            self.rechterNachfolger = self.rechterNachfolger.Einfuegen(datenNeu)
        return self

    def Suchen(self, suchSchluessel):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return self.daten
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.Suchen(suchSchluessel)
        else:
            return self.rechterNachfolger.Suchen(suchSchluessel)

    def IstVorhanden(self, suchSchluessel):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return True
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.IstVorhanden(suchSchluessel)
        else:
            return self.rechterNachfolger.IstVorhanden(suchSchluessel)

    def HöheGeben(self, aktTiefe):
        linkeHöhe = self.linkerNachfolger.HöheGeben(aktTiefe + 1)
        rechteHöhe = self.rechterNachfolger.HöheGeben(aktTiefe + 1)
        return max(linkeHöhe, rechteHöhe)

    def TiefeGeben(self, suchSchluessel, tiefe):
        if self.daten.SchluesselIstGleich(suchSchluessel):
            return tiefe
        elif self.daten.SchluesselIstGroesserAls(suchSchluessel):
            return self.linkerNachfolger.TiefeGeben(suchSchluessel, tiefe + 1)
        else:
            return self.rechterNachfolger.TiefeGeben(suchSchluessel, tiefe + 1)
        

    def KnotenZeichnen(self, xmin, xmax, y, dy, symbolVorgänger):
    
        self.knotenSymbol = KnotenSymbol(x=0, y=0, radius=8, farbe="schwarz", bezeichner=self.daten.WortGeben() )
        self.knotenSymbol.PositionSetzen(xmin + (xmax - xmin) / 2, y)
        # Eine Kante wird nur gezeichnet, wenn es einen Vorgänger gibt (d.h. dieser Knoten nicht die Wurzel ist)
        if symbolVorgänger != None:
            self.kantenSymbol =  KantenSymbol(symbolVorgänger, self.knotenSymbol, False, "", 3, "schwarz");            

        self.rechterNachfolger.KnotenZeichnen(xmin + (xmax - xmin) / 2, xmax, y + dy, dy, self.knotenSymbol)
        self.linkerNachfolger.KnotenZeichnen(xmin, xmin + (xmax - xmin) / 2, y + dy, dy, self.knotenSymbol)
    


class BinBaum:
    """
    Die Klasse Binbaum ist die Grundstruktur eines geordneten Binärbaums.
    Konkret wird hier der Binärbaum genutzt um ein einfaches Wörterbuch umzusetzen,
    deshalb sind die Daten von der Klasse Woerterbucheintrag.
    Die Methoden die ein Objekt dieser Klasse anbietet, leiten den Aufruf an den
    Wurzelknoten weiter, wenn der Baum nicht leer ist.
    (Implementierung mit Entwurfsmuster Kompositum)
    """
    def __init__(self):
        """
        Konstruktor fuer Objekte der Klasse Binbaum
        Ein leerer Baum wird erzeugt: Er besteht nur aus einem Abschluss-Objekt.
        """
        self.wurzel = Abschluss()

    def Einfuegen(self, datenNeu):
        """
        Fuegt ein Datenelement sortiert in den geordneten Binärbaum ein,
        d.h. die Eigenschaft, dass der Baum geordnet ist bleibt bei jedem
        Einfuegevorgang erhalten.
        :param datenNeu: neues Datenelement
        """
        self.wurzel = self.wurzel.Einfuegen(datenNeu)

        # Am Ende muss noch das Zeichnen des Baumes angestoßen werden
        # Normalerweise sollte das Zeichnen des Baumes explizit angestoßen werden,
        # aus Gründen der Einfachheit wird es hier aber direkt nach dem Einfügen aufgerufen
        self.BaumZeichnen()

    def Suchen(self, suchSchluessel):
        """
        Sucht ein Datenelement, das ueber seinen Schluessel
        identifiziert wird, und gibt es aus.
        :param suchSchluessel: Schluessel, nach dem gesucht wird
        :return: gesuchtes Datenelement bzw. None, falls die Suche erfolglos ist
        """
        return self.wurzel.Suchen(suchSchluessel)

    def IstVorhanden(self, suchSchluessel):
        """
        ueberprueft, ob ein Datenelement mit dem eingegebenen Schluessel (in Form einer
        Zeichenkette) vorhanden ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: True, im Erfolgsfall, False sonst.
        """
        return self.wurzel.IstVorhanden(suchSchluessel)

    def HöheGeben(self):
        """
        Berechnet die Höhe des Baums und gibt diesen Wert zurueck
        :return: Höhe des Baums 
        """
        return self.wurzel.HöheGeben(0)

    def TiefeGeben(self, suchSchluessel):
        """
        Gibt die Tiefe des Knotens an, in dem ein Datenelement mit dem Schluessel suchSchluessel gespeichert ist.
        :param suchSchluessel: Schluessel, nach dem gesucht wird als Zeichenkette 
        :return: Tiefe als positive Zahl, bzw. -1 falls der suchSchluessel nicht vorhanden ist.
        """
        return self.wurzel.TiefeGeben(suchSchluessel, 0)

    def BaumZeichnen(self):
        höhenschritt = 40
        breite = 800

        # y= 20 fuer die 0te Ebene, damit der Knoten sichtbar ist
        self.wurzel.KnotenZeichnen(0, breite,  25, höhenschritt, None) 



woerterb1 = Woerterbucheintrag("clip", "Clip, Spange, Klammer")
woerterb2 = Woerterbucheintrag("cat", "Katze")
woerterb3 = Woerterbucheintrag("cup", "Tasse, Becher")
woerterb4 = Woerterbucheintrag("cut", "Schnitt, Öffnung, kürzen")
woerterb5 = Woerterbucheintrag("cube", "Würfel")
woerterb6 = Woerterbucheintrag("cave", "Höhle")
binBaum1 = BinBaum()
binBaum1.Einfuegen(woerterb1)
binBaum1.Einfuegen(woerterb2)
binBaum1.Einfuegen(woerterb3)
binBaum1.Einfuegen(woerterb4)
